###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import pytest
import watertap.property_models.cryst_prop_pack as props
from pyomo.environ import ConcreteModel, SolverFactory, TerminationCondition
from idaes.core import FlowsheetBlock, ControlVolume0DBlock
from idaes.generic_models.properties.tests.test_harness import \
    PropertyTestHarness as PropertyTestHarness_idaes
from watertap.property_models.tests.property_test_harness import \
    (PropertyTestHarness, PropertyRegressionTest, PropertyCalculateStateTest)


# -----------------------------------------------------------------------------

class TestNaClProperty_idaes(PropertyTestHarness_idaes):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}
        self.prop_args = {}
        self.has_density_terms = False

class TestDefaultNaClwaterProperty:

    # Create block and stream for running default tests
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = props.NaClParameterBlock(default=
                                           {"heat_of_crystallization_model": props.HeatOfCrystallizationModel.constant})
    m.fs.stream = m.fs.properties.build_state_block([0], default={'defined_state': True})

    m.fs.cv = ControlVolume0DBlock(default={
        "dynamic": False,
        "has_holdup": False,
        "property_package": m.fs.properties})
    m.fs.cv.add_state_blocks(
        has_phase_equilibrium=False)
    m.fs.cv.add_material_balances()
    m.fs.cv.add_energy_balances()
    m.fs.cv.add_momentum_balances()


    # Create instance of PropertyTesthARNESS class and add attributes needed for tests
    xv = PropertyTestHarness()

    xv.stateblock_statistics = {'number_variables': 43,
                                  'number_total_constraints': 37,
                                  'number_unused_variables': 0, 
                                  'default_degrees_of_freedom': 6} 


    xv.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e0,
                         ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e1,
                         ('flow_mol_phase_comp', ('Sol', 'NaCl')): 1e2,
                         ('flow_mol_phase_comp', ('Vap', 'H2O')): 1e3,
                         }


    xv.default_solution = {('solubility_mass_phase_comp', ('Liq', 'NaCl')): 359.51,
                            ('solubility_mass_frac_phase_comp', ('Liq', 'NaCl')): 0.265,
                            ('dens_mass_solvent', 'Liq'): 996.89,
                            ('dens_mass_solvent', 'Vap') : 0.7363,
                            ('dens_mass_solute', 'Liq') : 3199.471,
                            ('dens_mass_solute', 'Sol') : 2115,
                            ('dens_mass_phase', 'Liq'): 1021.50,
                            ('dh_vap_solvent', None): 2441.80,
                            ('cp_solvent', 'Liq'): 4186.52,
                            ('cp_solvent', 'Vap'): 1864.52,
                            ('cp_solute', 'Sol'): 864.15, # cp_solute liquid ignored for now
                            ('cp_phase', 'Liq'): 4008.30,
                            ('flow_vol_phase', 'Liq'): 9.79e-4,
                            ('flow_vol_phase', 'Sol'): 0, 
                            ('flow_vol_phase', 'Vap'): 0,
                            ('pressure_sat', None): 2932.43,
                            ('temperature_sat_solvent', None): 296.79,
                            ('conc_mass_phase_comp', ('Liq', 'NaCl')): 35.753,
                            ('conc_mass_phase_comp', ('Liq', 'H2O')): 985.753, 
                            ('enth_mass_solvent', 'Liq'):  104.9212,
                            ('enth_mass_solvent', 'Vap'): 2546.7289,
                            ('enth_mass_solute', 'Sol'): 21.4739,
                            ('enth_mass_phase', 'Liq'): 101.0922,
                            ('dh_crystallization', 'NaCl'): -520,
                            ('mass_frac_phase_comp', ('Liq', 'H2O')) : 0.965,
                            ('mass_frac_phase_comp', ('Liq', 'NaCl')) : 0.035,
                            ('mass_frac_phase_comp', ('Sol', 'NaCl')) : 1.0,
                            ('mass_frac_phase_comp', ('Vap', 'H2O')) : 1.0,
                            ('flow_mol_phase_comp', ('Liq', 'H2O')) : 53.57,
                            ('flow_mol_phase_comp', ('Liq', 'NaCl')) : 0.5989,
                            ('flow_mol_phase_comp', ('Sol', 'NaCl')) : 1.6318503332497833e-09,
                            ('flow_mol_phase_comp', ('Vap', 'H2O')) : 1.6318424528638692e-07,
                            ('mole_frac_phase_comp', ('Liq', 'H2O')) : 0.9889,
                            ('mole_frac_phase_comp', ('Liq', 'NaCl')) : 0.01106,
                            ('mole_frac_phase_comp', ('Sol', 'NaCl')) : 1.0,
                            ('mole_frac_phase_comp', ('Vap', 'H2O')) : 1.0,
                            }

    # Configure class
    xv.configure_class(m)


    @pytest.mark.unit
    def test_components_phases(self):
        self.xv.test_components_phases(self.m)

    @pytest.mark.unit
    def test_parameters(self):
        self.xv.test_parameters(self.m)

    @pytest.mark.unit
    def test_state_variables(self):
        self.xv.test_state_variables(self.m)

    @pytest.mark.unit
    def test_permanent_properties(self):
        self.xv.test_permanent_properties(self.m)

    @pytest.mark.unit
    def test_on_demand_properties(self):
        self.xv.test_on_demand_properties(self.m)

    @pytest.mark.unit
    def test_stateblock_statistics(self):
        self.xv.test_stateblock_statistics(self.m)

    @pytest.mark.unit
    def test_units_consistent(self):
        self.xv.test_units_consistent(self.m)

    @pytest.mark.unit
    def test_scaling(self):
        self.xv.test_scaling(self.m)

    @pytest.mark.component
    def test_default_initialization(self):
        self.xv.test_default_initialization(self.m)

    @pytest.mark.component
    def test_property_control_volume(self):
        self.xv.test_property_control_volume(self.m)

@pytest.mark.component
class TestNaClPropertySolution_1(PropertyRegressionTest):
    # Test pure liquid solution 1 - same solution as NaCl prop pack
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e0,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e2,
                             }

        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.95,
                           ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0.05,
                           ('flow_mass_phase_comp', ('Sol', 'NaCl')): 0.0,
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 0.0,
                           ('temperature', None): 273.15 + 25,
                           ('pressure', None): 50e5}

        self.regression_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.95,
                                    ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                                    ('solubility_mass_phase_comp', ('Liq', 'NaCl')) : 359.50,
                                    ('solubility_mass_frac_phase_comp', ('Liq', 'NaCl')) : 0.265,
                                    ('dens_mass_solvent', 'Liq'): 996.89,
                                    ('dens_mass_solvent', 'Vap'): 36.335,
                                    ('dens_mass_phase', 'Liq'): 1032.2,
                                    ('flow_vol_phase', 'Liq'): 9.687e-4,
                                    ('conc_mass_phase_comp', ('Liq', 'H2O')): 980.6,
                                    ('conc_mass_phase_comp', ('Liq', 'NaCl')): 51.61,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 52.73,
                                    ('flow_mol_phase_comp', ('Liq', 'NaCl')): 0.8556,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9840,
                                    ('mole_frac_phase_comp', ('Liq', 'NaCl')): 1.597e-2,
                                    ('cp_solvent', 'Liq'): 4186.52,
                                    ('cp_phase', 'Liq'): 3940.03,
                                    ('enth_mass_solvent', 'Liq'): 104.92,
                                    ('enth_mass_phase', 'Liq'): 99.45
                                    }

@pytest.mark.component
class TestNaClPropertySolution_2(PropertyRegressionTest):
    # Test pure liquid solution 2 - same solution as NaCl prop pack
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e0,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e2,
                             }

        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.74,
                           ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0.26,
                           ('flow_mass_phase_comp', ('Sol', 'NaCl')): 0.0,
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 0.0,
                           ('temperature', None): 273.15 + 25,
                           ('pressure', None): 50e5}

        self.regression_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.74,
                                    ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.26,
                                    ('solubility_mass_phase_comp', ('Liq', 'NaCl')) : 359.50,
                                    ('solubility_mass_frac_phase_comp', ('Liq', 'NaCl')) : 0.265,
                                    ('dens_mass_solvent', 'Liq'): 996.89,
                                    ('dens_mass_phase', 'Liq'): 1193.4,
                                    ('flow_vol_phase', 'Liq'): 8.379e-4,
                                    ('conc_mass_phase_comp', ('Liq', 'H2O')): 883.14,
                                    ('conc_mass_phase_comp', ('Liq', 'NaCl')): 310.29,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 41.08,
                                    ('flow_mol_phase_comp', ('Liq', 'NaCl')): 4.449,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9022,
                                    ('mole_frac_phase_comp', ('Liq', 'NaCl')): 9.773e-2,
                                    ('cp_solvent', 'Liq'): 4186.52,
                                    ('cp_phase', 'Liq'): 3276.56,
                                    ('enth_mass_solvent', 'Liq'): 104.92,
                                    ('enth_mass_phase', 'Liq'): 68.78}

@pytest.mark.component
class TestNaClPropertySolution_3(PropertyRegressionTest):
    # Test pure liquid solution 3 - same solution as NaCl prop pack
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e0,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e3,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e3,
                             }

        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.999,
                           ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0.001,
                           ('flow_mass_phase_comp', ('Sol', 'NaCl')): 0.0,
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 0.0,
                           ('temperature', None): 273.15 + 25,
                           ('pressure', None): 10e5}

        self.regression_solution = {('mass_frac_phase_comp', ('Liq', 'H2O')): 0.999,
                                    ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.001,
                                    ('solubility_mass_phase_comp', ('Liq', 'NaCl')) : 359.50,
                                    ('solubility_mass_frac_phase_comp', ('Liq', 'NaCl')) : 0.265,
                                    ('dens_mass_solvent', 'Liq'): 996.89,
                                    ('dens_mass_solvent', 'Vap'): 7.267,
                                    ('dens_mass_phase', 'Liq'): 997.59,
                                    ('flow_vol_phase', 'Liq'): 1.002e-3,
                                    ('conc_mass_phase_comp', ('Liq', 'H2O')): 996.59,
                                    ('conc_mass_phase_comp', ('Liq', 'NaCl')): 0.9976,
                                    ('flow_mol_phase_comp', ('Liq', 'H2O')): 55.45,
                                    ('flow_mol_phase_comp', ('Liq', 'NaCl')): 1.711e-2,
                                    ('mole_frac_phase_comp', ('Liq', 'H2O')): 0.9997,
                                    ('mole_frac_phase_comp', ('Liq', 'NaCl')): 3.084e-4,
                                    ('cp_solvent', 'Liq'): 4186.52,
                                    ('cp_phase', 'Liq'): 4180.98,
                                    ('enth_mass_solvent', 'Liq'): 104.92,
                                    ('enth_mass_phase', 'Liq'): 104.40}

@pytest.mark.component
class TestNaClPropertySolution_4(PropertyRegressionTest):
    # Test pure solid solution 1 - check solid properties
    def configure(self):

        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e3,
                             ('flow_mass_phase_comp', ('Sol', 'NaCl')): 1e0,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e2
                             }

        self.state_args = {('flow_vol_phase', 'Liq'): 0,
                           ('flow_vol_phase', 'Vap'): 0,
                           ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                           ('flow_mass_phase_comp', ('Sol', 'NaCl')): 1.0,
                           ('temperature', None): 273.15 + 25,
                           ('pressure', None): 5e5}

        self.regression_solution = {('mass_frac_phase_comp', ('Sol', 'NaCl')): 1.0,
                                    ('dens_mass_solute', 'Sol'): 2115,
                                    ('cp_solute', 'Sol'): 864.16,
                                    ('flow_vol_phase', 'Sol'): 1 / 2115, # mass floe / density
                                    ('cp_solute', 'Sol'): 864.16,
                                    ('enth_mass_solute', 'Sol'): 21.474,
                                    ('dh_crystallization', 'NaCl'): -520,                                   
                                    ('flow_mol_phase_comp', ('Sol', 'NaCl')): 1 / 58.44E-3, # mass flow / mw
                                    ('mole_frac_phase_comp', ('Sol', 'NaCl')): 1.0,
                                    }

@pytest.mark.component
class TestNaClPropertySolution_5(PropertyRegressionTest):
    # Test pure vapor solution 1 - check vapor properties
    def configure(self):

        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e3,
                             ('flow_mass_phase_comp', ('Sol', 'NaCl')): 1e3,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e0
                             }

        self.state_args = {('flow_vol_phase', 'Liq'): 0,
                           ('flow_vol_phase', 'Sol'): 0,
                           ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 1.0,
                           ('temperature', None): 273.15 + 25,
                           ('pressure', None): 5e5}

        self.regression_solution = {('mass_frac_phase_comp', ('Vap', 'H2O')): 1.0,
                                    ('dens_mass_solvent', 'Vap'): 3.633,
                                    ('dh_vap_solvent', None): 2441.808,
                                    ('cp_solvent', 'Vap'): 1864.52,
                                    ('flow_vol_phase', 'Vap'): 1 / 3.633, # mass flow / density
                                    ('pressure_sat', None): 2905.28,                             
                                    ('flow_mol_phase_comp', ('Vap', 'H2O')): 1 / 18.01528E-3, # mass flow / mw
                                    ('mole_frac_phase_comp', ('Vap', 'H2O')): 1.0,
                                    }

@pytest.mark.component
class TestNaClPropertySolution_6(PropertyRegressionTest):
    # Test for S-L-V solution 1 with similar magnitude flowrates in all phases and high liquid salt. conc. - check all properties
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Sol', 'NaCl')): 10,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e1
                             }

        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.75,
                           ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0.25,
                           ('flow_mass_phase_comp', ('Sol', 'NaCl')): 0.25,
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 0.25,
                           ('temperature', None): 273.15 + 50,
                           ('pressure', None): 5e5}

        self.regression_solution = {('solubility_mass_phase_comp', ('Liq', 'NaCl')): 362.93,
                                ('solubility_mass_frac_phase_comp', ('Liq', 'NaCl')): 0.2686,
                                ('dens_mass_solvent', 'Liq'): 988.04,
                                ('dens_mass_solvent', 'Vap') : 3.352,
                                ('dens_mass_solute', 'Liq') : 2645.21,
                                ('dens_mass_solute', 'Sol') : 2115,
                                ('dens_mass_phase', 'Liq'): 1171.53,
                                ('dh_vap_solvent', None): 2382.08,
                                ('cp_solvent', 'Liq'): 4180.92,
                                ('cp_solvent', 'Vap'): 1871.21,
                                ('cp_solute', 'Sol'): 873.34,
                                ('cp_phase', 'Liq'): 3300.83,
                                ('flow_vol_phase', 'Liq'): 8.53e-4,
                                ('flow_vol_phase', 'Sol'): 1.182e-4, 
                                ('flow_vol_phase', 'Vap'): 7.46e-2,
                                ('pressure_sat', None): 9799.91,
                                ('temperature_sat_solvent', None): 318.52,
                                ('conc_mass_phase_comp', ('Liq', 'NaCl')): 292.88,
                                ('conc_mass_phase_comp', ('Liq', 'H2O')): 878.65, 
                                ('enth_mass_solvent', 'Liq'):  209.40,
                                ('enth_mass_solvent', 'Vap'): 2591.48,
                                ('enth_mass_solute', 'Sol'): 43.19,
                                ('enth_mass_phase', 'Liq'): 152.36,
                                ('dh_crystallization', 'NaCl'): -520,
                                ('mass_frac_phase_comp', ('Liq', 'H2O')) : 0.75,
                                ('mass_frac_phase_comp', ('Liq', 'NaCl')) : 0.25,
                                ('mass_frac_phase_comp', ('Sol', 'NaCl')) : 1,
                                ('mass_frac_phase_comp', ('Vap', 'H2O')) : 1,
                                ('flow_mol_phase_comp', ('Liq', 'H2O')) : 41.63,
                                ('flow_mol_phase_comp', ('Liq', 'NaCl')) : 4.28,
                                ('flow_mol_phase_comp', ('Sol', 'NaCl')) : 4.28,
                                ('flow_mol_phase_comp', ('Vap', 'H2O')) : 13.88,
                                ('mole_frac_phase_comp', ('Liq', 'H2O')) : 0.9068,
                                ('mole_frac_phase_comp', ('Liq', 'NaCl')) : 0.0932,
                                ('mole_frac_phase_comp', ('Sol', 'NaCl')) : 1.0,
                                ('mole_frac_phase_comp', ('Vap', 'H2O')) : 1.0,
                                }

@pytest.mark.component
class TestNaClPropertySolution_7(PropertyRegressionTest):
    # Test for S-L-V solution 2 with flowrates in all phases of same magnitude but low liquid salt. conc. - check all properties
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Sol', 'NaCl')): 10,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e3
                             }

        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 0.999,
                           ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0.001,
                           ('flow_mass_phase_comp', ('Sol', 'NaCl')): 0.25,
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 0.25,
                           ('temperature', None): 273.15 + 50,
                           ('pressure', None): 5e5}

        self.regression_solution = {('solubility_mass_phase_comp', ('Liq', 'NaCl')): 362.93,
                                ('solubility_mass_frac_phase_comp', ('Liq', 'NaCl')): 0.2686,
                                ('dens_mass_solvent', 'Liq'): 988.04,
                                ('dens_mass_solvent', 'Vap') : 3.352,
                                ('dens_mass_solute', 'Liq') : 3073.05,
                                ('dens_mass_solute', 'Sol') : 2115,
                                ('dens_mass_phase', 'Liq'): 988.72,
                                ('dh_vap_solvent', None): 2382.08,
                                ('cp_solvent', 'Liq'): 4180.92,
                                ('cp_solvent', 'Vap'): 1871.21,
                                ('cp_solute', 'Sol'): 873.34,
                                ('cp_phase', 'Liq'): 4175.95,
                                ('flow_vol_phase', 'Liq'): 1.011e-3,
                                ('flow_vol_phase', 'Sol'): 1.182e-4, 
                                ('flow_vol_phase', 'Vap'): 7.46e-2,
                                ('pressure_sat', None): 12614.93,
                                ('temperature_sat_solvent', None): 323.55,
                                ('conc_mass_phase_comp', ('Liq', 'NaCl')): 0.9887,
                                ('conc_mass_phase_comp', ('Liq', 'H2O')): 987.73, 
                                ('enth_mass_solvent', 'Liq'):  209.40,
                                ('enth_mass_solvent', 'Vap'): 2591.48,
                                ('enth_mass_solute', 'Sol'): 43.19,
                                ('enth_mass_phase', 'Liq'): 208.81,
                                ('dh_crystallization', 'NaCl'): -520,
                                ('mass_frac_phase_comp', ('Liq', 'H2O')) : 0.999,
                                ('mass_frac_phase_comp', ('Liq', 'NaCl')) : 0.001,
                                ('mass_frac_phase_comp', ('Sol', 'NaCl')) : 1,
                                ('mass_frac_phase_comp', ('Vap', 'H2O')) : 1,
                                ('flow_mol_phase_comp', ('Liq', 'H2O')) : 55.45,
                                ('flow_mol_phase_comp', ('Liq', 'NaCl')) : 0.0171,
                                ('flow_mol_phase_comp', ('Sol', 'NaCl')) : 4.28,
                                ('flow_mol_phase_comp', ('Vap', 'H2O')) : 13.88,
                                ('mole_frac_phase_comp', ('Liq', 'H2O')) : 0.999,
                                ('mole_frac_phase_comp', ('Liq', 'NaCl')) : 3.084e-4,
                                ('mole_frac_phase_comp', ('Sol', 'NaCl')) : 1.0,
                                ('mole_frac_phase_comp', ('Vap', 'H2O')) : 1.0,
                                }

@pytest.mark.component
class TestNaClPropertySolution_8(PropertyRegressionTest):
    # Test for S-L-V solution 3 with outlet data from Dutta et al. - proper crystallization system
    # # Dutta recorded properties for solids and liquids at crystallizer temperature:
    # # Solubility @ 55C: 0.27 kg/kg
    # # Heat of vaporization @ 55C: 2400 kJ/kg
    # # Vapor density @ 55C: 68.7e-3 kg/m3
    # # Solid heat capacity: 877 J/kgK
    # # Liquid heat capacity @ 20 C: 3290 kJ/kgK
    # # Liquid density @ 20 C : 1185 kg/m3

    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Sol', 'NaCl')): 1e0,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e0, 
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e-1, 

                             }

        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 18.37, # 84t/h
                           ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.2126,
                           ('flow_mass_phase_comp', ('Sol', 'NaCl')): 5.55, # 20t/h
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 20.55, # 74t/h
                           ('temperature', None): 273.15 + 20,
                           ('pressure', None): 10e3}

        self.regression_solution = {('solubility_mass_phase_comp', ('Liq', 'NaCl')): 358.88,
                                ('solubility_mass_frac_phase_comp', ('Liq', 'NaCl')): 0.2644,        
                                ('dens_mass_solvent', 'Liq'): 998.02,
                                ('dens_mass_solvent', 'Vap') : 73.91e-3,
                                ('dens_mass_solute', 'Liq') : 2847.63,
                                ('dens_mass_solute', 'Sol') : 2115,
                                ('dens_mass_phase', 'Liq'): 1157.91,
                                ('dh_vap_solvent', None): 2453.66,
                                ('cp_solvent', 'Liq'): 4189.43,
                                ('cp_solvent', 'Vap'): 1863.46,
                                ('cp_solute', 'Sol'): 862.16,
                                ('cp_phase', 'Liq'): 3380.07,
                                ('flow_vol_phase', 'Liq'): 0.02015,
                                ('flow_vol_phase', 'Sol'): 2.624e-3, 
                                ('flow_vol_phase', 'Vap'): 278.04,
                                ('pressure_sat', None): 1679.64,
                                ('temperature_sat_solvent', None): 287.88,
                                ('conc_mass_phase_comp', ('Liq', 'NaCl')): 246.17,
                                ('conc_mass_phase_comp', ('Liq', 'H2O')): 911.74, 
                                ('enth_mass_solvent', 'Liq'):  84.00,
                                ('enth_mass_solvent', 'Vap'): 2537.66,
                                ('enth_mass_solute', 'Sol'): 17.16,
                                ('enth_mass_phase', 'Liq'): 59.03,
                                ('dh_crystallization', 'NaCl'): -520,
                                ('mass_frac_phase_comp', ('Liq', 'H2O')) : 0.7874,
                                ('mass_frac_phase_comp', ('Sol', 'NaCl')) : 1,
                                ('mass_frac_phase_comp', ('Vap', 'H2O')) : 1,
                                ('flow_mol_phase_comp', ('Liq', 'H2O')) : 1019.69,
                                ('flow_mol_phase_comp', ('Liq', 'NaCl')) : 84.88,
                                ('flow_mol_phase_comp', ('Sol', 'NaCl')) : 94.97,
                                ('flow_mol_phase_comp', ('Vap', 'H2O')) : 1140.698,
                                ('mole_frac_phase_comp', ('Liq', 'H2O')) : 0.9232,
                                ('mole_frac_phase_comp', ('Liq', 'NaCl')) : 0.0768,
                                ('mole_frac_phase_comp', ('Sol', 'NaCl')) : 1.0,
                                ('mole_frac_phase_comp', ('Vap', 'H2O')) : 1.0,
                                ('flow_mass_phase_comp', ('Liq', 'NaCl')): 4.96,

                                }

@pytest.mark.component
class TestNaClPropertySolution_9(PropertyRegressionTest):
    # Test for S-L-V solution 4 with outlet data from Dutta et al. - proper crystallization system
    # # Dutta recorded properties for solids and liquids at crystallizer temperature:
    # # Solubility @ 55C: 0.27 kg/kg
    # # Heat of vaporization @ 55C: 2400 kJ/kg
    # # Vapor density @ 55C: 68.7e-3 kg/m3
    # # Solid heat capacity: 877 J/kgK
    # # Liquid heat capacity @ 20 C: 3290 kJ/kgK
    # # Liquid density @ 20 C : 1185 kg/m3

    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Sol', 'NaCl')): 1e0,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e0, 
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e-1, 

                             }

        self.state_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 18.37, # 84t/h
                           ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.2126,
                           ('flow_mass_phase_comp', ('Sol', 'NaCl')): 5.55, # 20t/h
                           ('flow_mass_phase_comp', ('Vap', 'H2O')): 20.55, # 74t/h
                           ('temperature', None): 273.15 + 55,
                           ('pressure', None): 10e3}

        self.regression_solution = {('solubility_mass_phase_comp', ('Liq', 'NaCl')): 363.71,
                                ('solubility_mass_frac_phase_comp', ('Liq', 'NaCl')): 0.2695,
                                ('dens_mass_solvent', 'Liq'): 985.71,
                                ('dens_mass_solvent', 'Vap') : 66.03e-3,
                                ('dens_mass_solute', 'Liq') : 2694.61,
                                ('dens_mass_solute', 'Sol') : 2115,
                                ('dens_mass_phase', 'Liq'): 1139.33,
                                ('dh_vap_solvent', None): 2369.98,
                                ('cp_solvent', 'Liq'): 4181.59,
                                ('cp_solvent', 'Vap'): 1872.79,
                                ('cp_solute', 'Sol'): 875.04,
                                ('cp_phase', 'Liq'): 3390.43,
                                ('flow_vol_phase', 'Liq'): 0.02047,
                                ('flow_vol_phase', 'Sol'): 2.624e-3, 
                                ('flow_vol_phase', 'Vap'): 311.233,
                                ('pressure_sat', None): 13201.77,
                                ('temperature_sat_solvent', None): 324.47,
                                ('conc_mass_phase_comp', ('Liq', 'NaCl')): 242.22,
                                ('conc_mass_phase_comp', ('Liq', 'H2O')): 897.107, 
                                ('enth_mass_solvent', 'Liq'):  230.30,
                                ('enth_mass_solvent', 'Vap'): 2600.279,
                                ('enth_mass_solute', 'Sol'): 47.57,
                                ('enth_mass_phase', 'Liq'): 177.39,
                                ('dh_crystallization', 'NaCl'): -520,
                                ('mass_frac_phase_comp', ('Liq', 'H2O')) : 0.7874,
                                ('mass_frac_phase_comp', ('Sol', 'NaCl')) : 1,
                                ('mass_frac_phase_comp', ('Vap', 'H2O')) : 1,
                                ('flow_mol_phase_comp', ('Liq', 'H2O')) : 1019.69,
                                ('flow_mol_phase_comp', ('Liq', 'NaCl')) : 84.88,
                                ('flow_mol_phase_comp', ('Sol', 'NaCl')) : 94.97,
                                ('flow_mol_phase_comp', ('Vap', 'H2O')) : 1140.698,
                                ('mole_frac_phase_comp', ('Liq', 'H2O')) : 0.9232,
                                ('mole_frac_phase_comp', ('Liq', 'NaCl')) : 0.0768,
                                ('mole_frac_phase_comp', ('Sol', 'NaCl')) : 1.0,
                                ('mole_frac_phase_comp', ('Vap', 'H2O')) : 1.0,
                                ('flow_mass_phase_comp', ('Liq', 'NaCl')): 4.96,

                                }

@pytest.mark.component
class TestNaClCalculateState_1(PropertyCalculateStateTest):
    # Test pure liquid solution with mass fractions
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e-1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e0,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1,
                             ('flow_mass_phase_comp', ('Sol', 'NaCl')): 1,
                             }

        self.var_args = {('flow_vol_phase', 'Liq'): 2e-2,
                         ('flow_vol_phase', 'Sol'): 0,
                         ('flow_vol_phase', 'Vap'): 0,
                         ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5
                         }

        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 19.6,
                               ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1.032,
                               ('flow_mass_phase_comp', ('Sol', 'NaCl')): 0,
                               ('flow_mass_phase_comp', ('Vap', 'H2O')): 0
                               }

@pytest.mark.component
class TestNaClCalculateState_2(PropertyCalculateStateTest):
    # Test pure liquid with mole fractions
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2,
                             # The rest are expected to be zero
                             ('flow_mass_phase_comp', ('Sol', 'NaCl')): 1,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1,
                             }
        self.var_args = {('flow_vol_phase', 'Liq'): 2e-4,
                         ('flow_vol_phase', 'Sol'): 0,
                         ('flow_vol_phase', 'Vap'): 0,
                         ('mole_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5}
        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 18.84e-2,
                               ('flow_mass_phase_comp', ('Liq', 'NaCl')): 3.215e-2,
                               ('flow_mass_phase_comp', ('Sol', 'NaCl')): 0,
                               ('flow_mass_phase_comp', ('Vap', 'H2O')): 0
                               }

@pytest.mark.component
class TestNaClCalculateState_3(PropertyCalculateStateTest):
    # Test pure liquid solution with pressure_sat defined instead of temperature
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e0,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e1,
                             ('flow_mass_phase_comp', ('Sol', 'NaCl')): 1e-1,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e2,
                             }

        self.var_args = {('flow_vol_phase', 'Liq'): 2e-2,
                         ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                         ('flow_vol_phase', 'Sol'): 0,
                         ('flow_vol_phase', 'Vap'): 0,
                         ('pressure_sat', None):  2905,
                         ('pressure', None): 5e5
                         }

        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 19.6,
                               ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1.032,
                               ('flow_mass_phase_comp', ('Sol', 'NaCl')): 0,
                               ('flow_mass_phase_comp', ('Vap', 'H2O')): 0,
                               ('temperature', None): 273.15 + 25
                               }

@pytest.mark.component
class TestNaClCalculateState_4(PropertyCalculateStateTest):
    # Test pure solid solution with mass fractions
    def configure(self):

        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
                             ('flow_mass_phase_comp', ('Sol', 'NaCl')): 1e-1,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e-1
                             }

        self.var_args = {('flow_vol_phase', 'Liq'): 0,
                         ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0,
                         ('flow_vol_phase', 'Sol'): 2e-2,
                         ('flow_vol_phase', 'Vap'): 0,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5
                         }

        self.state_solution = {('flow_mass_phase_comp', ('Sol', 'NaCl')): 2115 * 2e-2, # solid density is constant
                               ('flow_mass_phase_comp', ('Liq', 'H2O')):  0,
                               ('flow_mass_phase_comp', ('Liq', 'NaCl')): 0,
                               ('flow_mass_phase_comp', ('Vap', 'H2O')):  0 
                               }

@pytest.mark.component
class TestNaClCalculateState_5(PropertyCalculateStateTest):
    # Test solid-liquid-vapor mixture solution with mass fractions
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e-1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e1,
                             ('flow_mass_phase_comp', ('Sol', 'NaCl')): 1e-2}

        self.var_args = {('flow_vol_phase', 'Liq'): 2e-2,
                         ('mass_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                         ('flow_vol_phase', 'Sol'): 2e-2,
                         ('flow_vol_phase', 'Vap'): 2e-2,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5
                         }

        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')):  19.6,
                               ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1.032,
                               ('flow_mass_phase_comp', ('Vap', 'H2O')):  0.07265, 
                               ('flow_mass_phase_comp', ('Sol', 'NaCl')): 42.3, 
                               }

@pytest.mark.component
class TestNaClCalculateState_6(PropertyCalculateStateTest):
    # Test liquid-solid-vapor mixture with mole fractions
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {('flow_mass_phase_comp', ('Liq', 'H2O')): 1e1,
                             ('flow_mass_phase_comp', ('Liq', 'NaCl')): 1e2,
                             ('flow_mass_phase_comp', ('Sol', 'NaCl')): 1e2,
                             ('flow_mass_phase_comp', ('Vap', 'H2O')): 1e3,
                             }

        self.var_args = {('flow_vol_phase', 'Liq'): 2e-4,
                         ('flow_vol_phase', 'Sol'): 2e-4,
                         ('flow_vol_phase', 'Vap'): 2e-4,
                         ('mole_frac_phase_comp', ('Liq', 'NaCl')): 0.05,
                         ('temperature', None): 273.15 + 25,
                         ('pressure', None): 5e5}

        self.state_solution = {('flow_mass_phase_comp', ('Liq', 'H2O')): 18.84e-2,
                               ('flow_mass_phase_comp', ('Liq', 'NaCl')): 3.215e-2,
                               ('flow_mass_phase_comp', ('Sol', 'NaCl')): 0.423,
                               ('flow_mass_phase_comp', ('Vap', 'H2O')): 3.632 * 2e-4 # Density from ideal gas law * vol. flow
                               }




###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for loading water source definitions
"""
import pytest
import os

from pyomo.environ import units
from pyomo.util.check_units import assert_units_equivalent

from watertap.core import Database

dbpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")

db = Database()

exclude_files = [
    "water_sources.yaml", "component_list.yaml", "default_case_study.yaml"]

tech_list = []
for f in os.listdir(dbpath):
    filename = os.fsdecode(f)
    if filename.endswith(".yaml") and filename not in exclude_files:
        tech_list.append(filename[:-5])


@pytest.mark.integration
@pytest.mark.parametrize("tech", tech_list)
def test_unit_parameter_files(tech):
    data = db._get_technology(tech)

    # Check that data has as default key
    assert "default" in data

    # Iterate overall entries in tech data and check for expected contents
    # TODO : Need to check up on this once everything is done
    pass_through = ["blending_reservoir",
                    "buffer_tank",
                    "chemical_addition",
                    "cooling_supply",
                    "co2_addition",
                    "energy_recovery",
                    "feed_water_tank",
                    "injection_well_disposal",
                    "intrusion_mitigation",
                    "landfill",
                    "municipal_drinking",
                    "municipal_wwtp",
                    "pump",
                    "storage_tank",
                    "static_mixer",
                    "smp",
                    "surface_discharge",
                    "sw_onshore_intake",
                    "tramp_oil_tank",
                    "water_pumping_station",
                    "well_field",
                   ]

    siso_full_recovery = ["uv_aop", "uv", "fixed_bed", "decarbonator", "chlorination"]

    no_energy_electric_flow_vol_inlet = ["constructed_wetlands",
                                         "energy_recovery",
                                         "gas_sparged_membrane",
                                         "mbr_denitrification",
                                         "mbr_nitrification",
                                         "multi_stage_bubble_aeration",
                                         "tri_media_filtration",
                                         "cartridge_filtration_with_backflush",
                                         "landfill",
                                         "well_field",
                                         "ion_exchange",
                                         "ozone_aop",
                                         "fixed_bed",
                                         "holding_tank",
                                         "heap_leaching",
                                         "nuclear_cooling_tower",
                                         "lime_softening",
                                         "ozonation",
                                         "cooling_tower",
                                         "gac",
                                         "tri_media_filtration_with_backflush",
                                         "sedimentation",
                                         "backwash_solids_handling",
                                         "ph_decrease",
                                         "ph_increase",
                                         "co2_addition",
                                         "coag_and_floc",
                                         "crystallizer",
                                         "iron_and_manganese_removal",
                                         "fluidized_bed",
                                         "surface_discharge",
                                         "solution_distribution_and_recovery_plant",
                                         "chemical_addition",
                                         "cartridge_filtration",
                                         "injection_well",
                                         "sw_onshore_intake",
                                         "filter_press",
                                         "municipal_drinking",
                                         "gac_pressure_30_min",
                                         "packed_tower_aeration",
                                         "treated_storage",
                                         "gac_gravity_60_min",
                                         "evaporation_pond",
                                         "lime_addition",
                                         "brine_concentrator",
                                         "agglom_stacking",
                                         "storage_tank",
                                         "evaporation_pond",
                                         "deep_well_injection",
                                         "water_pumping_station",
                                         ]

    expected = ["recovery_frac_mass_H2O",
                "default_removal_frac_mass_solute"]

    for k in data.values():

        for e in expected:
            if tech not in pass_through and tech not in siso_full_recovery:
                assert e in k.keys()
                assert "units" in k[e].keys()
                assert_units_equivalent(
                    k[e]["units"], units.dimensionless)
                assert "value" in k[e].keys()
                assert k[e]["value"] >= 0
                assert k[e]["value"] <= 1
            elif tech in siso_full_recovery:
                if e == "default_removal_frac_mass_solute":
                    assert e in k.keys()
            else:
                assert e not in k.keys()

        if tech not in no_energy_electric_flow_vol_inlet:
            e = "energy_electric_flow_vol_inlet"
            assert e in k.keys()
            assert "units" in k[e].keys()
            assert_units_equivalent(
                k[e]["units"], units.dimensionless)
            assert "value" in k[e].keys()
            assert k[e]["value"] >= 0

        # Check for specific removal fractions
        if "removal_frac_mass_solute" in k.keys():
            for (j, c_data) in k["removal_frac_mass_solute"].items():
                assert "units" in c_data.keys()
                assert_units_equivalent(
                    c_data["units"], units.dimensionless)
                assert "value" in c_data.keys()
                assert c_data["value"] >= 0
                assert c_data["value"] <= 1
                assert j in db.component_list.keys()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order granular activated carbon model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    Block, check_optimal_termination, ConcreteModel, Constraint, value, Var)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import GACZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestGACZO_w_o_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["tss", "nonvolatile_toc"]})

        m.fs.unit = GACZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.empty_bed_contact_time, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.activated_carbon_replacement, Var)
        assert isinstance(model.fs.unit.activated_carbon_demand, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.electricity_intensity_constraint,
                          Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)
        assert isinstance(model.fs.unit.activated_carbon_equation, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("gac")

        model.fs.unit.load_parameters_from_database()
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.empty_bed_contact_time.fixed
        assert model.fs.unit.empty_bed_contact_time.value == \
            data["empty_bed_contact_time"]["value"]/60  # min to hour
        assert model.fs.unit.electricity_intensity_parameter.fixed
        assert model.fs.unit.electricity_intensity_parameter.value == \
            data["electricity_intensity_parameter"]["value"]
        assert model.fs.unit.activated_carbon_replacement.fixed
        assert model.fs.unit.activated_carbon_replacement.value == \
            data["activated_carbon_replacement"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(9.60083, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.00312473, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.083326, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(2.41793, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.49854, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(637.988, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                              : Value    : Fixed : Bounds
             Activated Carbon Demand :   1983.6 : False : (0, None)
                  Electricity Demand :   637.99 : False : (0, None)
               Electricity Intensity : 0.017718 : False : (None, None)
              Empty Bed Contact Time :  0.16667 :  True : (0, None)
    Solute Removal [nonvolatile_toc] :  0.20000 :  True : (0, None)
                Solute Removal [tss] :  0.97000 :  True : (0, None)
                      Water Recovery :  0.96000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                         Inlet    Treated  Byproduct
    Volumetric Flowrate                  10.002    9.6008  0.40117  
    Mass Concentration H2O               999.80    999.91   997.08  
    Mass Concentration tss             0.099980 0.0031247   2.4179  
    Mass Concentration nonvolatile_toc 0.099980  0.083326  0.49854  
====================================================================================
"""

        assert output in stream.getvalue()


class TestGACZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["tss", "nonvolatile_toc", "foo"]})

        m.fs.unit = GACZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.empty_bed_contact_time, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.electricity_intensity_constraint,
                          Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("gac")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == 0
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.empty_bed_contact_time.fixed
        assert model.fs.unit.empty_bed_contact_time.value == \
            data["empty_bed_contact_time"]["value"]/60  # min to hour
        assert model.fs.unit.electricity_intensity_parameter.fixed
        assert model.fs.unit.electricity_intensity_parameter.value == \
            data["electricity_intensity_parameter"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(9.60183, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.0031244, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.083317, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(0.1041468, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(2.41793, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.49854, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(2.4927e-08, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]))
        assert (pytest.approx(638.051, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                              : Value    : Fixed : Bounds
             Activated Carbon Demand :   1983.8 : False : (0, None)
                  Electricity Demand :   638.05 : False : (0, None)
               Electricity Intensity : 0.017718 : False : (None, None)
              Empty Bed Contact Time :  0.16667 :  True : (0, None)
                Solute Removal [foo] :   0.0000 :  True : (0, None)
    Solute Removal [nonvolatile_toc] :  0.20000 :  True : (0, None)
                Solute Removal [tss] :  0.97000 :  True : (0, None)
                      Water Recovery :  0.96000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                         Inlet    Treated  Byproduct
    Volumetric Flowrate                  10.003    9.6018    0.40117
    Mass Concentration H2O               999.70    999.81     997.08
    Mass Concentration tss             0.099970 0.0031244     2.4179
    Mass Concentration nonvolatile_toc 0.099970  0.083317    0.49854
    Mass Concentration foo             0.099970   0.10415 2.4927e-08
====================================================================================
"""
        assert output in stream.getvalue()


db = Database()
params = db._get_technology("gac")


class TestGACZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["tss", "nonvolatile_toc"]})

        m.fs.unit = GACZO(default={
            "property_package": m.fs.params,
            "database": db})

        return m

    @pytest.mark.parametrize("subtype", [params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("gac", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]


db = Database()
params = db._get_technology("gac")


@pytest.mark.component
@pytest.mark.parametrize("subtype", [k for k in params.keys()])
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = GACZO(default={
        "property_package": m.fs.params,
        "database": m.db,
        "process_subtype": subtype})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.gac, Block)
    assert isinstance(m.fs.costing.gac.capital_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.gac.capital_b_parameter,
                      Var)
    assert isinstance(m.fs.costing.gac.capital_c_parameter,
                      Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]
    assert m.fs.unit1.activated_carbon_demand[0] in \
        m.fs.costing._registered_flows["activated_carbon"]



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

# Import Pyomo libraries
from pyomo.environ import (Block,
                           Set,
                           Var,
                           Param,
                           Suffix,
                           NonNegativeReals,
                           Reference,
                           units as pyunits)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault,
                        MaterialFlowBasis)
from idaes.core.util import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

@declare_process_block_class("Condenser")
class CompressorData(UnitModelBlockData):
    """
    Condenser model for MVC
    """
    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The filtration unit does not support dynamic
    behavior, thus this must be False."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The filtration unit does not have defined volume, thus
    this must be False."""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}"""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
        **default** - MaterialBalanceType.useDefault.
        **Valid values:** {
        **MaterialBalanceType.useDefault - refer to property package for default
        balance type
        **MaterialBalanceType.none** - exclude material balances,
        **MaterialBalanceType.componentPhase** - use phase component balances,
        **MaterialBalanceType.componentTotal** - use total component balances,
        **MaterialBalanceType.elementTotal** - use total element balances,
        **MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
        **default** - EnergyBalanceType.useDefault.
        **Valid values:** {
        **EnergyBalanceType.useDefault - refer to property package for default
        balance type
        **EnergyBalanceType.none** - exclude energy balances,
        **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
        **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
        **EnergyBalanceType.energyTotal** - single energy balance for material,
        **EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))

    def build(self):
        super().build()

        # this creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Next, get the base units of measurement from the property definition
        units_meta = self.config.property_package.get_metadata().get_derived_units

        # Add control volume
        self.control_volume = ControlVolume0DBlock(default={
            "dynamic": False,
            "has_holdup": False,
            "property_package": self.config.property_package,
            "property_package_args": self.config.property_package_args})

        self.control_volume.add_state_blocks(
            has_phase_equilibrium=False)

        # complete condensation mass balance
        @self.control_volume.Constraint(self.flowsheet().time,
                                        self.config.property_package.component_list,
                                        doc="Mass balance")
        def mass_balance(b, t, j):
            lb = b.properties_out[t].get_material_flow_terms('Vap', j).lb
            b.properties_out[t].get_material_flow_terms('Vap', j).fix(lb)
            return (b.properties_in[t].get_material_flow_terms('Vap', j)
                    + b.properties_in[t].get_material_flow_terms('Liq', j)
                    == b.properties_out[t].get_material_flow_terms('Liq', j))

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=True)

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type)

        # # Add constraints
        @self.Constraint(self.flowsheet().time, doc="Saturation pressure constraint")
        def eq_condenser_pressure_sat(b, t):
            return (b.control_volume.properties_out[t].pressure
                    >= b.control_volume.properties_out[t].pressure_sat)

        # Add ports
        self.add_port(name='inlet', block=self.control_volume.properties_in)
        self.add_port(name='outlet', block=self.control_volume.properties_out)

    def initialize(
            blk,
            state_args=None,
            outlvl=idaeslog.NOTSET,
            solver=None,
            optarg=None,
            hold_state=False):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)
            hold_state: boolean indicating if the inlet conditions should stay fixed

        Returns: None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize control volume
        flags = blk.control_volume.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Initialization Step 1 Complete.")
        # # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high(
            "Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        if hold_state:
            return flags
        else:
            # Release Inlet state
            blk.control_volume.release_state(flags, outlvl=outlvl)
            init_log.info(
                "Initialization Complete: {}".format(idaeslog.condition(res))
            )

    def _get_performance_contents(self, time_point=0):
        var_dict = {"Heat duty": self.control_volume.heat[time_point]}
        return {"vars": var_dict}

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for (t, j), c in self.control_volume.mass_balance.items():
            sf = iscale.get_scaling_factor(
                self.control_volume.properties_in[t].flow_mass_phase_comp['Vap', j])
            iscale.constraint_scaling_transform(c, sf)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains utility functions for infeasibility diagnostics of WaterTAP models.
"""
import logging, sys

from contextlib import contextmanager
from math import isclose
from pyomo.environ import Var, Constraint, value
from pyomo.util.infeasible import log_infeasible_constraints, log_infeasible_bounds

_logger = logging.getLogger('watertap.core.util.infeasible.print')
_logger.setLevel(logging.DEBUG)

@contextmanager
def _logging_handler(output_file):
    if output_file is None:
        _handler = logging.StreamHandler(stream=sys.stdout)
    else:
        _handler = logging.FileHandler(output_file, mode='w')
    _handler.setFormatter(logging.Formatter('%(message)s'))
    _logger.addHandler(_handler)
    try:
        yield _logger
    finally:
        _logger.removeHandler(_handler)

def print_infeasible_constraints(m, tol=1e-6, print_expression=False,
        print_variables=False, output_file=None):
    """
    print the infeasble constraints in the model
    """
    with _logging_handler(output_file) as logger:
        log_infeasible_constraints(m, tol=tol, logger=logger,
                log_expression=print_expression, log_variables=print_variables)

def print_infeasible_bounds(m, tol=1e-6, output_file=None):
    """
    print the infeasible variable bounds in the model
    """
    with _logging_handler(output_file) as logger:
        log_infeasible_bounds(m, tol=tol, logger=logger)

def print_variables_close_to_bounds(m, rel_tol=1e-4, abs_tol=1e-12):
    """
    print variables close to their bounds
    """
    for var in m.component_data_objects(ctype=Var, descend_into=True):
        if var.fixed:
            continue
        val = var.value
        if val is None:
            print(f"No value for Var {var.name}")
        else:
            _eval_close(var, val, rel_tol, abs_tol)

def print_constraints_close_to_bounds(m, rel_tol=1e-4, abs_tol=1e-5):
    """
    print constraints close to their bounds
    """
    for con in m.component_data_objects(ctype=Constraint,
            descend_into=True, active=True):
        if con.equality:
            continue
        val = value(con.body, exception=False)
        if val is None:
            print(f"Cannot evaluate Constraint {con.name}: missing variable value")
        else:
            _eval_close(con, val, rel_tol, abs_tol)

def print_close_to_bounds(m, rel_tol=1e-04, abs_tol=1e-12):
    """
    Print variables and constraints which are near their bounds
    """
    print_variables_close_to_bounds(m, rel_tol=rel_tol, abs_tol=abs_tol)
    print_constraints_close_to_bounds(m, rel_tol=rel_tol, abs_tol=abs_tol)


def _eval_close(obj, val, rel_tol, abs_tol):
    lb = value(obj.lb); ub = value(obj.ub)
    if lb is not None and ub is not None and \
            isclose(lb, ub, rel_tol=rel_tol, abs_tol=abs_tol):
        return
    if lb is not None and isclose(lb, val, rel_tol=rel_tol, abs_tol=abs_tol):
        print(f"{obj.name} near LB of {lb}")
    if ub is not None and isclose(val, ub, rel_tol=rel_tol, abs_tol=abs_tol):
        print(f"{obj.name} near UB of {ub}")



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of cofermentation
for wastewater resource recovery flowsheets.
"""

from idaes.core import declare_process_block_class
from watertap.core import build_diso, ZeroOrderBaseData, constant_intensity
from pyomo.environ import Var, Constraint, units as pyunits

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("CofermentationZO")
class CofermentationZOData(ZeroOrderBaseData):
    """
    Zero-Order model for cofermentation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "cofermentation"


        #TODO: consider making a diso_reactive build function and adding conditional for
        # cod/nonbiodegradable cod to be in solute set. For now, unit assumes any solutes provided are
        # cod with a removal fraction to get the final ffCOD (i.e., (1-removal_frac)*mass_cod_in = mass_ffCOD

        if "cod" not in self.config.property_package.solute_set:
            raise ValueError('cod must be included in the solute list since'
                             ' this unit model converts cod to nonbiodegradable_cod.')
        build_diso(self)
        constant_intensity(self)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order landfill model
"""
import pytest
from io import StringIO

from pyomo.environ import Block, ConcreteModel, Constraint, value, Var
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import LandfillZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestLandfillZOdefault:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["sulfur", "toc", "tss"]})

        m.fs.unit = LandfillZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1e-5)
        m.fs.unit.inlet.flow_mass_comp[0, "sulfur"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(20)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(30)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == 'landfill'
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.capacity_basis, Var)
        assert isinstance(model.fs.unit.total_mass, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("landfill")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

        assert model.fs.unit.capacity_basis[0].fixed
        assert model.fs.unit.capacity_basis[0].value == data[
            "capacity_basis"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        for t, j in model.fs.unit.inlet.flow_mass_comp:
            assert (pytest.approx(value(
                model.fs.unit.inlet.flow_mass_comp[t, j]), rel=1e-5) ==
                value(model.fs.unit.outlet.flow_mass_comp[t, j]))
        assert (pytest.approx(216000.036, abs=1e-5) ==
                value(model.fs.unit.total_mass[0]))
        assert (pytest.approx(0.0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                    : Value      : Fixed : Bounds
    Capacity Basis (kg/hr) : 1.0000e+05 :  True : (None, None)
        Electricity Demand : 1.6544e-24 : False : (0, None)
     Electricity Intensity :     0.0000 :  True : (None, None)
        Total Mass (kg/hr) : 2.1600e+05 : False : (None, None)

------------------------------------------------------------------------------------
    Stream Table
                                 Inlet     Outlet  
    Volumetric Flowrate         0.060000   0.060000
    Mass Concentration H2O    0.00016667 0.00016667
    Mass Concentration sulfur     166.67     166.67
    Mass Concentration toc        333.33     333.33
    Mass Concentration tss        500.00     500.00
====================================================================================
"""

        assert output == stream.getvalue()

db = Database()
params = db._get_technology("landfill")

class TestLandfillZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["sulfur", "toc", "tss"]})

        m.fs.unit = LandfillZO(default={
            "property_package": m.fs.params,
            "database": db})

        return m

    @pytest.mark.parametrize("subtype", [params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters(
            "landfill", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

        assert model.fs.unit.capacity_basis[0].fixed
        assert model.fs.unit.capacity_basis[0].value == data[
            "capacity_basis"]["value"]


db = Database()
params = db._get_technology("landfill")

@pytest.mark.parametrize("subtype", [k for k in params.keys()])
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.params = WaterParameterBlock(
            default={"solute_list": ["sulfur", "toc", "tss"]})
    m.fs.costing = ZeroOrderCosting()
    m.fs.unit = LandfillZO(default={
                    "property_package": m.fs.params,
                    "database": m.db,
                    "process_subtype": subtype})

    m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1e-5)
    m.fs.unit.inlet.flow_mass_comp[0, "sulfur"].fix(1000)
    m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(2000)
    m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(3000)

    m.fs.unit.load_parameters_from_database()

    m.fs.unit.costing = UnitModelCostingBlock(default={
                    "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.landfill, Block)
    assert isinstance(m.fs.costing.landfill.capital_a_parameter, Var)
    assert isinstance(m.fs.costing.landfill.capital_b_parameter, Var)

    assert isinstance(m.fs.unit.costing.capital_cost, Var)
    assert isinstance(m.fs.unit.costing.capital_cost_constraint,
                      Constraint)
    assert_units_consistent(m.fs)

    assert degrees_of_freedom(m.fs.unit) == 0

    initialization_tester(m)
    _ = solver.solve(m)

    assert isinstance(m.fs.unit.costing.capital_cost_constraint,
                            Constraint)

    if subtype == "default":
        assert (pytest.approx(43.5627, rel=1e-5) ==
                value(m.fs.unit.costing.capital_cost))
    if subtype == "landfill_zld":
        assert (pytest.approx(20.09155, rel=1e-5) ==
                value(m.fs.unit.costing.capital_cost))

    assert m.fs.unit.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order Chlorination model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    Block, check_optimal_termination, ConcreteModel, Constraint, value, Var)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import ChlorinationZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestChlorinationZO_with_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(default={
            "solute_list": ["total_coliforms_fecal_ecoli",
                            "viruses_enteric",
                            "tss"]})

        m.fs.unit = ChlorinationZO(default={"property_package": m.fs.params,
                                            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(100)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == "chlorination"
        assert isinstance(model.fs.unit.initial_chlorine_demand, Var)
        assert isinstance(model.fs.unit.contact_time, Var)
        assert isinstance(model.fs.unit.concentration_time, Var)
        assert isinstance(model.fs.unit.chlorine_decay_rate, Var)
        assert isinstance(model.fs.unit.chlorine_dose, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.chlorine_dose_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("chlorination")
        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j not in data["removal_frac_mass_solute"]:
                assert v.value == data[
                    "default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.initial_chlorine_demand[0].fixed
        assert model.fs.unit.initial_chlorine_demand[0].value == data[
            "initial_chlorine_demand"]["value"]
        assert model.fs.unit.contact_time[0].fixed
        assert model.fs.unit.contact_time[0].value == data[
            "contact_time"]["value"]
        assert model.fs.unit.concentration_time[0].fixed
        assert model.fs.unit.concentration_time[0].value == data[
            "concentration_time"]["value"]
        assert model.fs.unit.chlorine_decay_rate[0].fixed
        assert model.fs.unit.chlorine_decay_rate[0].value == data[
            "chlorine_decay_rate"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.101001, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.008435544, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "total_coliforms_fecal_ecoli"]))
        assert (pytest.approx(0.00332669354, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "viruses_enteric"]))
        assert (pytest.approx(9.900873, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp[
                    "tss"]))
        assert (pytest.approx(9.5, rel=1e-5) ==
                value(model.fs.unit.chlorine_dose[0]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                          : Value      : Fixed : Bounds
                           CT Value ((mg*min)/L) :     450.00 :  True : (None, None)
                 Chlorine Decay Rate (mg/(L*hr)) :     3.0000 :  True : (None, None)
                            Chlorine Dose (mg/L) :     9.5000 : False : (None, None)
                               Contact Time (hr) :     1.5000 :  True : (None, None)
                              Electricity Demand :   0.018540 : False : (0, None)
                           Electricity Intensity : 5.0000e-05 :  True : (None, None)
                  Initial Chlorine Demand (mg/L) :     0.0000 :  True : (None, None)
    Solute Removal [total_coliforms_fecal_ecoli] :    0.99915 :  True : (0, None)
                            Solute Removal [tss] :     0.0000 :  True : (0, None)
                Solute Removal [viruses_enteric] :    0.99966 :  True : (0, None)

------------------------------------------------------------------------------------
    Stream Table
                                                     Inlet   Treated
    Volumetric Flowrate                            0.10300   0.10100
    Mass Concentration H2O                          970.87    990.09
    Mass Concentration total_coliforms_fecal_ecoli  9.7087 0.0084355
    Mass Concentration viruses_enteric              9.7087 0.0033267
    Mass Concentration tss                          9.7087    9.9009
====================================================================================
"""

        assert output in stream.getvalue()


class TestChlorinationZO_w_o_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(default={
            "solute_list": ["total_coliforms_fecal_ecoli", 
                            "viruses_enteric"]})

        m.fs.unit = ChlorinationZO(default={"property_package": m.fs.params,
                                            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(100)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == "chlorination"
        assert isinstance(model.fs.unit.initial_chlorine_demand, Var)
        assert isinstance(model.fs.unit.contact_time, Var)
        assert isinstance(model.fs.unit.concentration_time, Var)
        assert isinstance(model.fs.unit.chlorine_decay_rate, Var)
        assert isinstance(model.fs.unit.chlorine_dose, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.chlorine_dose_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("chlorination")
        model.fs.unit.load_parameters_from_database()
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j not in data["removal_frac_mass_solute"]:
                assert v.value == data[
                    "default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.initial_chlorine_demand[0].fixed
        assert model.fs.unit.initial_chlorine_demand[0].value == data[
            "initial_chlorine_demand"]["value"]
        assert model.fs.unit.contact_time[0].fixed
        assert model.fs.unit.contact_time[0].value == data[
            "contact_time"]["value"]
        assert model.fs.unit.concentration_time[0].fixed
        assert model.fs.unit.concentration_time[0].value == data[
            "concentration_time"]["value"]
        assert model.fs.unit.chlorine_decay_rate[0].fixed
        assert model.fs.unit.chlorine_decay_rate[0].value == data[
            "chlorine_decay_rate"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.100001, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.0085198987836, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "total_coliforms_fecal_ecoli"]))
        assert (pytest.approx(0.003359960083, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "viruses_enteric"]))
        assert (pytest.approx(9.5, rel=1e-5) ==
                value(model.fs.unit.chlorine_dose[0]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                          : Value      : Fixed : Bounds
                           CT Value ((mg*min)/L) :     450.00 :  True : (None, None)
                 Chlorine Decay Rate (mg/(L*hr)) :     3.0000 :  True : (None, None)
                            Chlorine Dose (mg/L) :     9.5000 : False : (None, None)
                               Contact Time (hr) :     1.5000 :  True : (None, None)
                              Electricity Demand :   0.018360 : False : (0, None)
                           Electricity Intensity : 5.0000e-05 :  True : (None, None)
                  Initial Chlorine Demand (mg/L) :     0.0000 :  True : (None, None)
    Solute Removal [total_coliforms_fecal_ecoli] :    0.99915 :  True : (0, None)
                Solute Removal [viruses_enteric] :    0.99966 :  True : (0, None)

------------------------------------------------------------------------------------
    Stream Table
                                                     Inlet   Treated
    Volumetric Flowrate                            0.10200   0.10000
    Mass Concentration H2O                          980.39    999.99
    Mass Concentration total_coliforms_fecal_ecoli  9.8039 0.0085199
    Mass Concentration viruses_enteric              9.8039 0.0033600
====================================================================================
"""

        assert output in stream.getvalue()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = ChlorinationZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.chlorination, Block)
    assert isinstance(m.fs.costing.chlorination.capital_a_parameter, Var)
    assert isinstance(m.fs.costing.chlorination.capital_b_parameter, Var)
    assert isinstance(m.fs.costing.chlorination.capital_c_parameter, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]
    assert str(m.fs.costing._registered_flows["chlorine"][0]) == str(
        m.fs.unit1.chlorine_dose[0]*m.fs.unit1.properties_in[0].flow_vol)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

# Import Pyomo libraries
from pyomo.environ import (Block,
                           Set,
                           Var,
                           Param,
                           Suffix,
                           NonNegativeReals,
                           Reference,
                           units as pyunits)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault,
                        MaterialFlowBasis)
from idaes.core.util import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog


_log = idaeslog.getLogger(__name__)

@declare_process_block_class("Compressor")
class CompressorData(UnitModelBlockData):
    """
    Compressor model for MVC
    """
    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The filtration unit does not support dynamic
    behavior, thus this must be False."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The filtration unit does not have defined volume, thus
    this must be False."""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}"""))
    CONFIG.declare("material_balance_type", ConfigValue( # TODO: update default material balance type
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
        **default** - MaterialBalanceType.useDefault.
        **Valid values:** {
        **MaterialBalanceType.useDefault - refer to property package for default
        balance type
        **MaterialBalanceType.none** - exclude material balances,
        **MaterialBalanceType.componentPhase** - use phase component balances,
        **MaterialBalanceType.componentTotal** - use total component balances,
        **MaterialBalanceType.elementTotal** - use total element balances,
        **MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
        **default** - EnergyBalanceType.useDefault.
        **Valid values:** {
        **EnergyBalanceType.useDefault - refer to property package for default
        balance type
        **EnergyBalanceType.none** - exclude energy balances,
        **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
        **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
        **EnergyBalanceType.energyTotal** - single energy balance for material,
        **EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))


    def build(self):

        super().build()

        # this creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Next, get the base units of measurement from the property definition
        units_meta = self.config.property_package.get_metadata().get_derived_units

        # Add unit variables
        self.pressure_ratio = Var(
            initialize=1.5,
            bounds=(1,10),
            units=pyunits.dimensionless
        )

        self.efficiency = Var(
            initialize=0.8,
            bounds=(1e-8,1),
            units=pyunits.dimensionless
        )

        # Build control volume
        self.control_volume = ControlVolume0DBlock(default={
            "dynamic": False,
            "has_holdup": False,
            "property_package": self.config.property_package,
            "property_package_args": self.config.property_package_args})

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=MaterialBalanceType.componentPhase,
            has_mass_transfer=False)

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_work_transfer=True)

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=True)

        # Add isentropic outlet block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = False # isentropic outlet is not an inlet
        self.properties_isentropic_out = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of isentropic outlet",
            default=tmp_dict)

        # Add ports - oftentimes users interact with these rather than the state blocks
        self.add_port(name='inlet', block=self.control_volume.properties_in)
        self.add_port(name='outlet', block=self.control_volume.properties_out)

        # References for control volume
        # pressure change
        # if (self.config.has_pressure_change is True and
        #         self.config.momentum_balance_type != 'none'):
        #     self.deltaP = Reference(self.control_volume.deltaP)

        # Add constraints
        # @self.Constraint(self.config.property_package.phase_list, doc="Mass balance for inlet/outlet")
        # def eq_mass_balance(b, p):
        #     return (b.control_volume.properties_in[0].flow_mass_phase_comp[p, 'H2O']
        #             == b.control_volume.properties_out[0].flow_mass_phase_comp[p, 'H2O'])

        @self.Constraint(self.config.property_package.phase_list, doc="Mass balance for inlet/isentropic outlet")
        def eq_mass_balance_isentropic(b, p):
            return (b.control_volume.properties_in[0].flow_mass_phase_comp[p, 'H2O']
                    == b.properties_isentropic_out[0].flow_mass_phase_comp[p, 'H2O'])

        @self.Constraint(doc="Pressure ratio")
        def eq_pressure_ratio(b):
            return b.control_volume.properties_in[0].pressure*b.pressure_ratio == b.control_volume.properties_out[0].pressure

        @self.Constraint(doc="Isentropic pressure")
        def eq_isentropic_pressure(b):
            return b.properties_isentropic_out[0].pressure == b.control_volume.properties_out[0].pressure

        @self.Constraint(doc="Isentropic temperature")
        def eq_isentropic_temperature(b):
            gamma = 1.3 # change to specific heat ratio
            return b.properties_isentropic_out[0].temperature == b.control_volume.properties_in[0].temperature * b.pressure_ratio ** (1-1/gamma)

        @self.Constraint(doc="Energy balance/efficiency definition")
        def eq_efficiency(b):
            return b.efficiency*(b.control_volume.properties_out[0].enth_mass_phase['Vap'] - b.control_volume.properties_in[0].enth_mass_phase['Vap'])\
                   == b.properties_isentropic_out[0].enth_mass_phase['Vap'] - b.control_volume.properties_in[0].enth_mass_phase['Vap']

    def initialize(
            blk,
            state_args=None,
            outlvl=idaeslog.NOTSET,
            solver=None,
            optarg=None):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize state blocks
        flags = blk.control_volume.initialize(solver=solver, optarg=optarg, hold_state=True)

        init_log.info_high("Initialization Step 1 Complete.")
        # ---------------------------------------------------------------------
        # Set state_args from inlet state
        if state_args is None:
            state_args = {}
            state_dict = blk.control_volume.properties_in[
                blk.flowsheet().config.time.first()].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        blk.properties_isentropic_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Initialization Step 2 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high(
            "Initialization Step 3 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.control_volume.release_state(flags, outlvl=outlvl)
        init_log.info(
            "Initialization Complete: {}".format(idaeslog.condition(res))
        )

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Pressure ratio"] = self.pressure_ratio
        var_dict["Efficiency"] = self.efficiency
        var_dict["Work"] = self.control_volume.work[time_point]

        return {"vars": var_dict}

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        iscale.set_scaling_factor(self.pressure_ratio, 1)
        iscale.set_scaling_factor(self.efficiency, 1)

        for j, c in self.eq_mass_balance_isentropic.items():
            sf = iscale.get_scaling_factor(self.control_volume.properties_in[0].flow_mass_phase_comp[j, 'H2O'])
            iscale.constraint_scaling_transform(c, sf)

        # Pressure constraints
        sf = iscale.get_scaling_factor(self.control_volume.properties_in[0].pressure)
        iscale.constraint_scaling_transform(self.eq_pressure_ratio, sf)
        iscale.constraint_scaling_transform(self.eq_isentropic_pressure, sf)

        # Temperature constraint
        sf = iscale.get_scaling_factor(self.control_volume.properties_in[0].temperature)
        iscale.constraint_scaling_transform(self.eq_isentropic_temperature, sf)

        # Efficiency, work constraints
        sf = iscale.get_scaling_factor(self.control_volume.work)
        iscale.constraint_scaling_transform(self.eq_efficiency, sf)


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order Ozonation model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    Block, check_optimal_termination, ConcreteModel, Constraint, value, Var)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import OzoneZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestOzoneZO_with_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(default={
            "solute_list": ["cryptosporidium",
                            "toc",
                            "giardia_lamblia",
                            "eeq",
                            "total_coliforms_fecal_ecoli",
                            "viruses_enteric",
                            "tss"]})

        m.fs.unit = OzoneZO(default={"property_package": m.fs.params,
                                     "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(100)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "giardia_lamblia"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "eeq"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)

        return m

    @pytest.mark.unit
    def test_toc_in_solute_list(self):
        model = ConcreteModel()
        model.db = Database()

        model.fs = FlowsheetBlock(default={"dynamic": False})
        model.fs.params = WaterParameterBlock(default={
            "solute_list": ["cryptosporidium",
                            "giardia_lamblia",
                            "eeq"]})
        with pytest.raises(ConfigurationError,
                           match="TOC must be in solute list for Ozonation "
                           "or Ozone/AOP"):
            model.fs.unit = OzoneZO(default={
                "property_package": model.fs.params,
                "database": model.db})

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert model.fs.unit._tech_type == "ozonation"
        assert isinstance(model.fs.unit.contact_time, Var)
        assert isinstance(model.fs.unit.concentration_time, Var)
        assert isinstance(model.fs.unit.mass_transfer_efficiency, Var)
        assert isinstance(model.fs.unit.ozone_flow_mass, Var)
        assert isinstance(model.fs.unit.ozone_consumption, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.specific_energy_coeff, Var)
        assert isinstance(model.fs.unit.ozone_consumption_constraint,
                          Constraint)
        assert isinstance(model.fs.unit.ozone_flow_mass_constraint,
                          Constraint)
        assert isinstance(model.fs.unit.electricity_constraint,
                          Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("ozonation")
        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j not in data["removal_frac_mass_solute"]:
                assert v.value == data[
                    "default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.contact_time[0].fixed
        assert model.fs.unit.contact_time[0].value == data[
            "contact_time"]["value"]
        assert model.fs.unit.concentration_time[0].fixed
        assert model.fs.unit.concentration_time[0].value == data[
            "concentration_time"]["value"]
        assert model.fs.unit.mass_transfer_efficiency[0].fixed
        assert model.fs.unit.mass_transfer_efficiency[0].value == data[
            "mass_transfer_efficiency"]["value"]
        assert model.fs.unit.specific_energy_coeff[0].fixed
        assert model.fs.unit.specific_energy_coeff[0].value == data[
            "specific_energy_coeff"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.102089, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(2.661333, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["toc"]))
        assert (pytest.approx(9.795299, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.103497, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["eeq"]))
        assert (pytest.approx(9921.863324, rel=1e-5) ==
                value(model.fs.unit.ozone_flow_mass[0]))
        assert (pytest.approx(49609.316620, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                          : Value   : Fixed : Bounds
                     Ozone CT Value ((mg*min)/L) :  1.0000 :  True : (None, None)
                        Ozone Contact Time (min) :  1.0000 :  True : (None, None)
                         Ozone Mass Flow (lb/hr) :  9921.9 : False : (0, None)
                  Ozone Mass Transfer Efficiency : 0.80000 :  True : (None, None)
                    Ozone Unit Power Demand (kW) :  49609. : False : (0, None)
                Solute Removal [cryptosporidium] : 0.29479 :  True : (0, None)
                            Solute Removal [eeq] : 0.98943 :  True : (0, None)
                Solute Removal [giardia_lamblia] : 0.90324 :  True : (0, None)
                            Solute Removal [toc] : 0.72830 :  True : (0, None)
    Solute Removal [total_coliforms_fecal_ecoli] : 0.99723 :  True : (0, None)
                            Solute Removal [tss] :  0.0000 :  True : (0, None)
                Solute Removal [viruses_enteric] : 0.99723 :  True : (0, None)

------------------------------------------------------------------------------------
    Stream Table
                                                     Inlet  Treated
    Volumetric Flowrate                            0.10700  0.10209
    Mass Concentration H2O                          934.58   979.53
    Mass Concentration cryptosporidium              9.3458   6.9078
    Mass Concentration toc                          9.3458   2.6613
    Mass Concentration giardia_lamblia              9.3458  0.94780
    Mass Concentration eeq                          9.3458  0.10350
    Mass Concentration total_coliforms_fecal_ecoli  9.3458 0.027182
    Mass Concentration viruses_enteric              9.3458 0.027182
    Mass Concentration tss                          9.3458   9.7953
====================================================================================
"""

        assert output in stream.getvalue()


class TestOzoneZO_w_o_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(default={
            "solute_list": ["cryptosporidium",
                            "toc",
                            "giardia_lamblia",
                            "eeq",
                            "total_coliforms_fecal_ecoli",
                            "viruses_enteric"]})

        m.fs.unit = OzoneZO(default={"property_package": m.fs.params,
                                     "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(100)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "giardia_lamblia"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "eeq"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert model.fs.unit._tech_type == "ozonation"
        assert isinstance(model.fs.unit.contact_time, Var)
        assert isinstance(model.fs.unit.concentration_time, Var)
        assert isinstance(model.fs.unit.mass_transfer_efficiency, Var)
        assert isinstance(model.fs.unit.ozone_flow_mass, Var)
        assert isinstance(model.fs.unit.ozone_consumption, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.specific_energy_coeff, Var)
        assert isinstance(model.fs.unit.ozone_consumption_constraint,
                          Constraint)
        assert isinstance(model.fs.unit.ozone_flow_mass_constraint, Constraint)
        assert isinstance(model.fs.unit.electricity_constraint,
                          Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("ozonation")
        model.fs.unit.load_parameters_from_database()
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j not in data["removal_frac_mass_solute"]:
                assert v.value == data[
                    "default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.contact_time[0].fixed
        assert model.fs.unit.contact_time[0].value == data[
            "contact_time"]["value"]
        assert model.fs.unit.concentration_time[0].fixed
        assert model.fs.unit.concentration_time[0].value == data[
            "concentration_time"]["value"]
        assert model.fs.unit.mass_transfer_efficiency[0].fixed
        assert model.fs.unit.mass_transfer_efficiency[0].value == data[
            "mass_transfer_efficiency"]["value"]
        assert model.fs.unit.specific_energy_coeff[0].fixed
        assert model.fs.unit.specific_energy_coeff[0].value == data[
            "specific_energy_coeff"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.101089, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(2.687660, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["toc"]))
        assert (pytest.approx(0.957178, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp[
                "giardia_lamblia"]))
        assert (pytest.approx(0.104520, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["eeq"]))
        assert (pytest.approx(9921.85340, rel=1e-5) ==
                value(model.fs.unit.ozone_flow_mass[0]))
        assert (pytest.approx(49609.267016, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                          : Value   : Fixed : Bounds
                     Ozone CT Value ((mg*min)/L) :  1.0000 :  True : (None, None)
                        Ozone Contact Time (min) :  1.0000 :  True : (None, None)
                         Ozone Mass Flow (lb/hr) :  9921.9 : False : (0, None)
                  Ozone Mass Transfer Efficiency : 0.80000 :  True : (None, None)
                    Ozone Unit Power Demand (kW) :  49609. : False : (0, None)
                Solute Removal [cryptosporidium] : 0.29479 :  True : (0, None)
                            Solute Removal [eeq] : 0.98943 :  True : (0, None)
                Solute Removal [giardia_lamblia] : 0.90324 :  True : (0, None)
                            Solute Removal [toc] : 0.72830 :  True : (0, None)
    Solute Removal [total_coliforms_fecal_ecoli] : 0.99723 :  True : (0, None)
                Solute Removal [viruses_enteric] : 0.99723 :  True : (0, None)

------------------------------------------------------------------------------------
    Stream Table
                                                     Inlet  Treated
    Volumetric Flowrate                            0.10600  0.10109
    Mass Concentration H2O                          943.40   989.22
    Mass Concentration cryptosporidium              9.4340   6.9761
    Mass Concentration toc                          9.4340   2.6877
    Mass Concentration giardia_lamblia              9.4340  0.95718
    Mass Concentration eeq                          9.4340  0.10452
    Mass Concentration total_coliforms_fecal_ecoli  9.4340 0.027451
    Mass Concentration viruses_enteric              9.4340 0.027451
====================================================================================
"""

        assert output in stream.getvalue()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = OzoneZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.ozonation, Block)
    assert isinstance(m.fs.costing.ozonation.ozone_capital_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.ozonation.ozone_capital_b_parameter,
                      Var)
    assert isinstance(m.fs.costing.ozonation.ozone_capital_c_parameter,
                      Var)
    assert isinstance(m.fs.costing.ozonation.ozone_capital_d_parameter,
                      Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]
